#!D:\bld\gdal-split_1642266023771\_h_env\python.exe
# EASY-INSTALL-SCRIPT: 'GDAL==3.3.3','gdal_sieve.py'
__requires__ = 'GDAL==3.3.3'
__import__('pkg_resources').run_script('GDAL==3.3.3', 'gdal_sieve.py')
